/*-
 * Copyright (C) @BABOLO  2002 Dec 08
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ifndef lint
static const char copyright[] = "\
@(#)Copyright (C) @BABOLO  2002 Dec 08\n\
@(#)All rights reserved.\n";
static const char rcsid[] = "$Id: binfout.c,v 1.2 2002/12/14 01:00:39 babolo Exp $";
#endif /* not lint */

#include <sysexits.h>
#include <stdlib.h>
#include <unistd.h>
#include <stdio.h>
#include <fcntl.h>
#include <err.h>
#include "bio_babolo.h"
#include "binfout.h"

int v_flag;
u_int64_t bin[FORMSIZE];

u_int64_t masks[] =
{ 0x5555555555555555, 0x3333333333333333, 0x0F0F0F0F0F0F0F0F
, 0x00FF00FF00FF00FF, 0x0000FFFF0000FFFF, 0x00000000FFFFFFFF
};
u_char shifts[] = {0x01, 0x02, 0x04, 0x08, 0x10, 0x20};

void
usage(int ex) {
    fprintf(stderr, "binfout " VERS "  " DATE "
Usage: binfout [-v[v]..] FormatFile
");
    exit(ex);
}

int
main(int argc, char *argv[]) {
    int in = fileno(stdin);
    FILE *out = stdout;
    FILE *fmt = NULL;
    int c;
    u_char *format, *buf;
    size_t i, j;
    struct tnode *root, *leaf;
    struct formades *inf;
    struct bio_ctrl *bio;

    v_flag = 0;
    while ((c = getopt(argc, argv, "qv?")) != -1) {
        switch (c) {
        case 'q': v_flag--;
                  break;
        case 'v': v_flag++;
                  break;
        case '?': usage(EX_OK);
        default:  usage(EX_USAGE);
    }   }
    argc -= optind;
    argv += optind;
    if  (argc < 1 || !argv || !*argv || !**argv) {
                            if  (v_flag >= 0) fprintf(stderr, "No FormatFile given\n"); exit(EX_USAGE);
    }
    if  (argc > 2)              {if (v_flag >= 0) fprintf(stderr, "Overaggumented\n"); exit(EX_USAGE);}
    if  ((fmt = fopen(*argv, "r")) == NULL) {
                            if (v_flag >= 0) err(EX_NOINPUT, "file %s:", *argv); else exit(EX_NOINPUT);
    }
    if  (argc == 2 && !argv[1] && !*argv[1]) {
                             if  (v_flag >= 0) fprintf(stderr, "No InputFile given\n"); exit(EX_USAGE);
    }
    if  (argc == 2 && (in = open(argv[1], O_RDONLY)) < 0) {
                          if (v_flag >= 0) err(EX_NOINPUT, "file %s:", argv[1]); else exit(EX_NOINPUT);
    }
    buf = format = bufread(fmt);
    inf = informat(&format);
    root = outformat(format);
    bio = bio_open(in, inf->tot >> 3);

    while ((buf = bio_get(bio, inf->tot >> 3)) != NULL) {
        for (i = 0; v_flag > 2 && i < (inf->tot >> 3); i++) fprintf(stderr, "%02X", buf[i]);
        if  (v_flag > 2) fprintf(stderr, "\n");
        for (i = 0, j = 0; i < inf->cur; i++) {
            for (bin[i] = 0, c = 0; c < (inf->form[i].ll >> 3); c++) {
                if  (v_flag > 3) fprintf(stderr, "i=%2d j=%3d c=%d f=%02X\n", i, j, c, buf[j]);
                bin[i] = (bin[i] << 8) | (buf[j++] & 0xFF);
            }
            if  (v_flag > 3) fprintf(stderr, "bin[%d]=%016llX\n", i, bin[i]);
            for (c = 5; c >= 0; c--) {
                if  (shifts[c] & inf->form[i].ff) {
                    if  (v_flag > 4)
                        fprintf( stderr, "c=%d %016llX %016llX %016llX"
                               , c, masks[c]
                               , ((bin[i] & masks[c]) << shifts[c])
                               , ((bin[i] >> shifts[c]) & masks[c])
                               );
                    bin[i] = ((bin[i] & masks[c]) << shifts[c]) | ((bin[i] >> shifts[c]) & masks[c]);
                    if  (v_flag > 4)
                        fprintf( stderr, " %016llX\nlen=%02X diff=%02X shift=%02X %02X"
                               , bin[i]
                               , inf->form[i].ll
                               , 64 - inf->form[i].ll
                               , shifts[c]
                               , (64 - inf->form[i].ll) & shifts[c]
                               );
                    if  ((64 - inf->form[i].ll) & shifts[c]) bin[i] = bin[i] >> shifts[c];
                    if  (v_flag > 4) fprintf(stderr, " %016llX\n", bin[i]);
            }   }
            if  (v_flag > 2) fprintf( stderr, "bin[%d]=%016llX\n", i, bin[i]);
        }
        for (leaf = root; leaf; leaf = leaf->next) {
            if  (v_flag > 3) {
                fprintf(stderr, "%2d %02X ~%s~", leaf->n, leaf->ff, leaf->f);
                if  (leaf->ff) fprintf(stderr, leaf->f, bin[leaf->n]);
                  else fprintf(stderr, leaf->f);
            }else{
                if  (leaf->ff) fprintf(out, leaf->f, bin[leaf->n]);
                  else fprintf(out, leaf->f);
    }   }   }
    bio_close(bio);
    exit(EX_OK);
}
